import random


def get_chronicle_name():
    firsts = [
        "super",
        "extra",
        "grand",
        "absolute",
        "joyful",
        "amazing",
        "relentless",
        "bold",
        "brave",
        "clever",
        "cool",
        "daring",
        "dreamy",
        "elegant",
        "epic",
        "fearless",
        "glorious",
        "heroic",
        "noble",
        "swift",
        "vibrant",
        "thunderous",
        "watching",
        "obversing",
    ]

    lasts = [
        "panda",
        "tiger",
        "hero",
        "challenger",
        "captain",
        "dragon",
        "eagle",
        "falcon",
        "fighter",
        "warrior",
        "knight",
        "leopard",
        "lion",
        "ninja",
        "runner",
        "ranger",
        "shadow",
        "sprinter",
        "observer",
        "watcher",
        "voyager",
        "starman",
        "wolf",
        "viking",
    ]

    chosen_first = random.randint(0, len(firsts) - 1)
    chosen_last = random.randint(0, len(lasts) - 1)

    return f"{firsts[chosen_first]}-{lasts[chosen_last]}"



import logging
import os

from colorama import Fore
import yaml

from .constants import EXAMPLE_YAML, README_CHRONICLES, README_WEIGHTS


class StartCommand:
    def __init__(self, project_type, project_name):
        self.project_name = project_name
        self.project_type = project_type
        
    def _get_project_pretty_type(self):
        pretty_types = {
            "classification": "Image Classification",
            "detection": "Object Detection"
        }
        
        return pretty_types[self.project_type]
    
    def create_datasets_structure(self):
        project_general_info = {"name": self.project_name, "type": self.project_type, "version": "0.1.0"}

        logging.info(
            f"Creating project structure for {Fore.CYAN}{self.project_name}{Fore.RESET}, a {Fore.CYAN}{self._get_project_pretty_type()}{Fore.RESET} project."
        )

        try:
            data = open("project.yaml", "r", encoding="utf8").read()
            if len(data) > 0:
                logging.info(
                    f"There is already a project created in this location. Use {Fore.CYAN}odin wrath{Fore.RESET} to delete the project, then run the command again."
                )
                return
        except:
            pass

        with open("project.yaml", "w", encoding="utf8") as wf:
            wf.write(yaml.dump(project_general_info))

        # Datasets
        dataset_parent = f"{os.path.abspath('.')}\\datasets"
        if not os.path.exists(dataset_parent):
            os.makedirs(dataset_parent)

            logging.info(f"Succesfully created {Fore.CYAN}datasets{Fore.RESET}")

        logging.info(f"Creating {Fore.CYAN}datasets examples{Fore.RESET}...")
    
    def create_models_structure(self):
        # Chronicles
        # chronicles/chronicle-nick/run-uuid
        chronicles_parent = f"{os.path.abspath('.')}\\chronicles"
        if not os.path.exists(chronicles_parent):
            os.makedirs(chronicles_parent)

            with open(f"{chronicles_parent}\\README.md", "w", encoding="utf8") as wf:
                wf.write(README_CHRONICLES)

            logging.info(f"Succesfully created {Fore.CYAN}chronicles{Fore.RESET}")

        # Weights
        weights_parent = f"{os.path.abspath('.')}\\weights"
        if not os.path.exists(weights_parent):
            os.makedirs(weights_parent)

            with open(f"{weights_parent}\\README.md", "w", encoding="utf8") as wf:
                wf.write(README_WEIGHTS)

            logging.info(f"Succesfully created {Fore.CYAN}weights{Fore.RESET}")
    
    def classification(self):
        dataset_example_classif = f"{os.path.abspath('.')}\\datasets\\dataset_1_classif"
        if not os.path.exists(dataset_example_classif):
            os.makedirs(dataset_example_classif)

            os.makedirs(f"{os.path.abspath('.')}\\datasets\\dataset_1_classif\\class_1")
            os.makedirs(f"{os.path.abspath('.')}\\datasets\\dataset_1_classif\\class_2")

            logging.info(
                f"Succesfully created dataset example at: {Fore.CYAN}datasets/dataset_1_classif{Fore.RESET}"
            )
            
    def detection(self):
        dataset_example_obd = f"{os.path.abspath('.')}\\datasets\\dataset_1_obd"
        if not os.path.exists(dataset_example_obd):
            os.makedirs(dataset_example_obd)

            os.makedirs(f"{os.path.abspath('.')}\\datasets\\dataset_1_obd\\train")
            os.makedirs(f"{os.path.abspath('.')}\\datasets\\dataset_1_obd\\val")
            os.makedirs(
                f"{os.path.abspath('.')}\\datasets\\dataset_1_obd\\train\\images"
            )
            os.makedirs(f"{os.path.abspath('.')}\\datasets\\dataset_1_obd\\val\\images")
            os.makedirs(
                f"{os.path.abspath('.')}\\datasets\\dataset_1_obd\\train\\labels"
            )
            os.makedirs(f"{os.path.abspath('.')}\\datasets\\dataset_1_obd\\val\\labels")

            with open(
                f"{os.path.abspath('.')}\\datasets\\dataset_1_obd\\data.yaml", "w"
            ) as wf:
                wf.write(EXAMPLE_YAML)

            logging.info(
                f"Succesfully created dataset example at: {Fore.CYAN}datasets/dataset_1_obd{Fore.RESET}"
            )


import logging
import os
import shutil
import subprocess
import uuid
import click
from colorama import Fore

from .chronicle_utils import get_chronicle_name

logging.basicConfig(
    level=logging.INFO, format=f"[{Fore.CYAN}ODIN{Fore.RESET}][{Fore.YELLOW}%(asctime)s{Fore.RESET}] %(message)s"
)

@click.group()
def cli():
    pass

@click.command("start")
@click.argument("project_name", default=None, required=False)
@click.option(
    "-t",
    "--type",
    "project_type",
    default=None,
    help="the project's type, must be either 'detection' (object detection) or 'classification'.",
)
def start(project_type, project_name):
    """Starts a new machine vision project."""
    from start import StartCommand
    
    if not project_name:
        confirmed_name = False
        
        while not confirmed_name:
            project_name = click.prompt(
                f"What is the project's name?"
            )
            
            if click.confirm(f'{Fore.CYAN}{project_name}{Fore.RESET}, is this name right?'):
                confirmed_name = True
    
    if not project_type:
        project_type = click.prompt(
            f"What is the project's type?",
            show_choices=True,
            type=click.Choice([f'classification', f'detection']),
        )
    
    builder = StartCommand(project_type, project_name)
    
    project_builders = {
        "classification": builder.classification,
        "detection": builder.detection
    }
    
    builder.create_datasets_structure()
    
    project_builders[project_type]()

    builder.create_models_structure()

@click.command("model")
@click.argument("action", type=click.Choice([
    'train',
    'test',
    'publish'
]))
@click.option(
    "--epochs",
    default=30,
    help="the number of epochs to train the model on, good values may differ from dataset to dataset.",
)
@click.option(
    "--device",
    default="cpu",
    help="the device to use to train the model, must be either 'cpu' or 'gpu'.",
)
@click.option(
    "--base_model",
    default=None,
    help="the pre-trained model to use for training.",
)
@click.option(
    "--subset",
    default=100,
    help="the percentage of the training dataset to use for the actual training of the model.",
)
@click.option(
    "-C",
    "--chronicle",
    "chronicle_name",
    default=None,
    help="the name of the chronicle to use."
)
@click.option(
    "-D",
    "--dataset",
    "dataset_name",
    default=None,
    help="the name of the dataset to use."
)
def model(action, epochs, device, base_model, subset, dataset_name, chronicle_name):
    """Trains the model, generating a new chronile based on a specific dataset. The name of the chronicle is not required, but can be passed."""
    # Command interface for model training.
    
    from project_utils import get_project_info
    project_info = get_project_info()
    project_type = project_info["type"]
                
    if project_type == "detection":
        from training_detection import DetectionTrainingCommands
        interpreter = DetectionTrainingCommands(project_type, dataset_name)
    elif project_type == "classification":
        from training_classification import ClassificationTrainingCommands
        interpreter = ClassificationTrainingCommands(project_type, dataset_name)
    
    commands = {
        "train": interpreter.train,
        "test": interpreter.test,
        "publish": interpreter.publish
    }
    
    commands[action](epochs=epochs, project_name=project_info["name"], device=device, base_model=base_model, dataset_name=dataset_name, chronicle_name=chronicle_name, subset=subset)

@click.command("dataset")
@click.argument("action", type=click.Choice([
    'create',
    'publish',
    'status',
    'delete', 
    'augmentate',
    'rollback', 
    'yaml'
]))
@click.option("-D", "--dataset_name", "dataset_name", default=None, help="the dataset name.")
@click.option(
    "-t",
    "--train",
    "train",
    default=70,
    help="represents the split of the training dataset",
)
@click.option(
    "-v",
    "--val",
    "val",
    default=30,
    help="represents the split of the validation dataset",
)
@click.option(
    "-a",
    "--augmentations",
    "augs",
    default=2,
    help="the amount of augmentations per image on augmentate command.",
)
@click.option(
    "-V",
    "--version",
    "rollver",
    default=None,
    help="the version to rollback to, must be a valid version.",
)
def dataset(action, dataset_name, train, val, augs, rollver):
    """Command interface for dataset management."""
    
    from dataset_classification import DatasetCommandsClassification
    from dataset_detection import DatasetCommandsDetection
    from project_utils import get_project_info
    
    project_info = get_project_info()
    project_type = project_info["type"]

    if not dataset_name:
        confirmed_name = False
        
        while not confirmed_name:
            dataset_name = click.prompt(
                f"What is the dataset's name?"
            )
            
            if click.confirm(f'{Fore.CYAN}{dataset_name}{Fore.RESET}, is this name right?'):
                confirmed_name = True

    if project_type == "detection":
        interpreter = DatasetCommandsDetection(dataset_name)
    elif project_type == "classification":
        interpreter = DatasetCommandsClassification(dataset_name)
        
    actions = {
        "create": interpreter.create,
        "publish": interpreter.publish,
        "status": interpreter.status,
        "delete": interpreter.delete,
        "augmentate": interpreter.augmentate,
        "rollback": interpreter.rollback,
        "yaml": interpreter.yaml
    }
    
    actions[action](train=train, val=val, augmentation_amount=augs, rollver=rollver)

@click.command("wrath")
def wrath():
    logging.info(
        f"Warning! Beyond this decision rests the {Fore.RED}DOOM{Fore.RESET} of all your {Fore.CYAN}datasets{Fore.RESET}, {Fore.CYAN}chronicles{Fore.RESET} and {Fore.CYAN}weights{Fore.RESET}."
    )

    if click.confirm("Do you want to continue?"):
        logging.info(
            f"Laying down {Fore.CYAN}Odin's{Fore.RESET} wrath against thy foe(der)s!"
        )

        try:
            shutil.rmtree("chronicles")
        except:
            pass

        try:
            shutil.rmtree("datasets")
        except:
            pass

        try:
            shutil.rmtree("weights")
        except:
            pass

        try:
            os.remove("project.yaml")
        except:
            pass

        logging.info("This land has been purged.")
    else:
        logging.info(f"So the {Fore.RED}Ragnarok{Fore.RESET} must wait.")


cli.add_command(start)
cli.add_command(wrath)
cli.add_command(dataset)
cli.add_command(model)

def main():
    cli()

if __name__ == "__main__":
    main()


import json
import logging
import os
import shutil

import click
from colorama import Fore


class BaseDatasetCommands:
    def __init__(self, type, dataset_name):
        self.type = type
        self.dataset_name = dataset_name
        self.dataset_path = f"{os.path.abspath('.')}\\datasets\\{dataset_name}"

    def _try_create_folder(self, folder_path):
        try:
            os.makedirs(folder_path)
        except FileExistsError:
            pass
        except Exception as e:
            logging.info(f"Something went wrong while trying to create {Fore.CYAN}{folder_path}{Fore.RESET}: {e}")

    def _upgrade_version(self, base_version, update_size):
        version = list(map(lambda x: int(x), base_version.split(".")))

        if update_size == "major":
            version[0] += 1
        elif update_size == "minor":
            version[1] += 1
        elif update_size == "fix":
            version[2] += 1

        upgraded_version = f"{version[0]}.{version[1]}.{version[2]}"

        return upgraded_version

    def _create_dataset_metadata_files(self, **kwargs):
        raise NotImplementedError(
            "This feature is not implemented or is going through maintenance."
        )

    def _publish_data(self, update_type="", train="", val="", **kwargs):
        raise NotImplementedError(
            "This feature is not implemented or is going through maintenance."
        )

    def _status_sum_staging(self, **kwargs):
        raise NotImplementedError(
            "This feature is not implemented or is going through maintenance."
        )

    def _status_sum_train(self, **kwargs):
        raise NotImplementedError(
            "This feature is not implemented or is going through maintenance."
        )

    def _status_show_additional_info(self, **kwargs):
        raise NotImplementedError(
            "This feature is not implemented or is going through maintenance."
        )

    def _augmentate_data(self, augs=0, **kwargs):
        raise NotImplementedError(
            "This feature is not implemented or is going through maintenance."
        )

    def _rollback_dataset(
        self, snapshot_info={}, rollback_version="", dataset_folder="", class_folder="", **kwargs
    ):
        raise NotImplementedError

    def create(self, **kwargs):
        logging.info(
            f"Creating dataset '{Fore.BLUE}{self.dataset_name}{Fore.RESET}'..."
        )

        if not os.path.exists(self.dataset_path):
            logging.info("Creating final folders...")

            self._create_dataset_metadata_files()
        else:
            logging.info(
                f"This dataset already exists, if you wish to recreate it or simply delete it, use the command {Fore.CYAN}odin dataset delete {self.dataset_name}{Fore.RESET}. This will delete the entire dataset, be careful."
            )

    def publish(self, train=0, val=0, **kwargs):
        logging.info(
            f"Publishing dataset '{Fore.BLUE}{self.dataset_name}{Fore.RESET}'..."
        )

        if not os.path.exists(self.dataset_path):
            logging.info(
                f"The dataset mentioned doesn't exist, create it by using the command {Fore.CYAN}odin dataset create {self.dataset_name}{Fore.RESET}."
            )
        else:
            update_type = click.prompt(
                f"Is this update to the dataset a {Fore.CYAN}fix{Fore.RESET}, {Fore.CYAN}minor{Fore.RESET} or {Fore.CYAN}major{Fore.RESET} update?",
                show_choices=True,
                type=click.Choice(["fix", "minor", "major"]),
            )

            self._publish_data(update_type=update_type, train=train, val=val)
            logging.info(f"Succesfully published {Fore.GREEN}train{Fore.RESET} data")

    def status(self, **kwargs):
        if not os.path.exists(self.dataset_path):
            logging.info(
                f"The dataset mentioned doesn't exist, create it by using the command {Fore.CYAN}odin dataset create {self.dataset_name}{Fore.RESET}."
            )
        else:
            status = {
                "empty": f"{Fore.YELLOW}Empty{Fore.RESET}",
                "staging": f"{Fore.YELLOW}Staging{Fore.RESET}",
                "published": f"{Fore.GREEN}Published{Fore.RESET}",
                "published_staging": f"{Fore.CYAN}Published and Staging{Fore.RESET}",
            }

            staging = self._status_sum_staging()
            train = self._status_sum_train()

            if staging > 0 and train > 0:
                logging.info(
                    f"Retrieving {Fore.CYAN}{self.dataset_name}{Fore.RESET} status..."
                )
                logging.info(f"Status: {status['published_staging']}")
            elif staging > 0 and train == 0:
                logging.info(
                    f"Retrieving {Fore.CYAN}{self.dataset_name}{Fore.RESET} status..."
                )
                logging.info(f"Status: {status['staging']}")
            elif train > 0 and staging == 0:
                logging.info(
                    f"Retrieving {Fore.CYAN}{self.dataset_name}{Fore.RESET} status..."
                )
                logging.info(f"Status: {status['published']}")
            else:
                logging.info(
                    f"Retrieving {Fore.CYAN}{self.dataset_name}{Fore.RESET} status..."
                )
                logging.info(f"Status: {status['empty']}")

            self._status_show_additional_info()

    def augmentate(self, augmentation_amount=0, **kwargs):
        if not os.path.exists(self.dataset_path):
            logging.info(
                f"The dataset mentioned doesn't exist, create it by using the command {Fore.CYAN}odin dataset create {self.dataset_name}{Fore.RESET}."
            )
        else:
            self._augmentate_data(augmentation_amount=augmentation_amount)

    def rollback(self, rollver="", **kwargs):
        if not os.path.exists(self.dataset_path):
            logging.info(
                f"The dataset mentioned doesn't exist, create it by using the command {Fore.CYAN}odin dataset create {self.dataset_name}{Fore.RESET}."
            )
        else:
            dataset_info = json.loads(
                open(f"{self.dataset_path}\\dataset.json", "r", encoding="utf8").read()
            )
            snapshot_info = json.loads(
                open(f"{self.dataset_path}\\snapshot.json", "r", encoding="utf8").read()
            )
            dataset_versions = []

            for version in snapshot_info:
                if version != dataset_info["version"]:
                    dataset_versions.append(version)

            if len(dataset_versions) > 0:
                if not rollver:
                    rollback_version = click.prompt(
                        f"Which version you want to rollback to?",
                        show_choices=True,
                        type=click.Choice(dataset_versions),
                    )

                    print(rollback_version)
                else:
                    if rollver.lower().replace("v", "") in dataset_versions:
                        rollback_version = rollver
                    else:
                        logging.info(
                            f"The version {Fore.LIGHTRED_EX}{rollver}{Fore.RESET} doesn't exist."
                        )
                        return

                # Add binaries from rollback version to actual version
                rollback_format = click.prompt(
                    f"Will the rollback replace the current version with the old one or create a new major version? (replace, create)",
                    show_choices=True,
                    type=click.Choice(["replace", "create"]),
                )

                if click.confirm(
                    "Do you want to continue? This action with replace all of your current files to a older version, the files might be deleted as well, this is not a merge."
                ):
                    if rollback_format == "replace":
                        final_version = rollback_version
                    else:
                        final_version = self._upgrade_version(
                            base_version=dataset_info["version"], update_size="major"
                        )

                    for dataset_folder in snapshot_info[rollback_version]:
                        for sub_folder in snapshot_info[rollback_version][
                            dataset_folder
                        ]:
                            self._rollback_dataset(
                                snapshot_info=snapshot_info,
                                rollback_version=rollback_version,
                                dataset_folder=dataset_folder,
                                class_folder=sub_folder,
                            )

                    dataset_info["version"] = final_version

                    if rollback_format == "create":
                        snapshot_info[final_version] = snapshot_info[rollback_version]

                    with open(
                        f"{self.dataset_path}\\dataset.json", "w", encoding="utf8"
                    ) as wf:
                        wf.write(json.dumps(dataset_info))

                    with open(
                        f"{self.dataset_path}\\snapshot.json", "w", encoding="utf8"
                    ) as wf:
                        wf.write(json.dumps(snapshot_info))

                    logging.info(
                        f"Succesfuly executed rollback to version {Fore.CYAN}{rollback_version}{Fore.RESET}, now at version {Fore.CYAN}{final_version}{Fore.RESET}"
                    )
            else:
                logging.info(
                    "There are no versions to rollback to, this dataset only have one snapshoted version, which is the current one."
                )

    def delete(self, **kwargs):
        dataset_path = f"{os.path.abspath('.')}\\datasets\\{self.dataset_name}"

        if not os.path.exists(dataset_path):
            logging.info(
                f"The dataset mentioned doesn't exist, create it by using the command {Fore.CYAN}odin dataset create {self.dataset_name}{Fore.RESET}."
            )
        else:
            if click.confirm(
                f"Do you want to continue? All your data will be lost (and you will bring {Fore.RED}The Ragnarok{Fore.RESET} to thy dataset!)"
            ):
                logging.info(
                    f"Aplying {Fore.RED}Ragnarok{Fore.RESET} to {Fore.CYAN}{self.dataset_name}{Fore.RESET}!"
                )

                try:
                    shutil.rmtree(f"{dataset_path}")

                    logging.info(
                        f"Successfully deleted {Fore.CYAN}{self.dataset_name}{Fore.RESET}."
                    )
                except:
                    logging.info(
                        f"{Fore.CYAN}Odin{Fore.RESET} was unable to delete {Fore.CYAN}{self.dataset_name}{Fore.RESET}."
                    )

    def yaml(self, **kwargs):
        raise NotImplementedError



import datetime
import json
import logging
import os
import shutil

from colorama import Fore
from .results import YOLOResult


class BaseTrainingCommands:
    def __init__(self, type, dataset_name):
        self.type = type
        self.dataset_name = dataset_name
        self.dataset_path = f"{os.path.abspath('.')}\\datasets\\{dataset_name}"
        
    def _get_chronicle_data(self, chronicle_path):
        with open(chronicle_path+"\\chronicle.json", "r", encoding="utf8") as f:
            data = json.loads(f.read())
        return data
        
    def _gather_yolo_results(self, results) -> list[YOLOResult]:
        odin_results = []
        
        for raw_result in results:
            result = YOLOResult(
                image_with_bboxes=raw_result.plot(),
                inference_speed=raw_result.speed
            )
            odin_results.append(result)
        
        return odin_results
        
    def _check_if_eligible_model(self, model_file_name):
        if model_file_name.startswith("epoch") or model_file_name in ["best.pt", "last.pt"]:
            return False
        return True
        
    def _try_create_folder(self, folder_path):
        try:
            os.makedirs(folder_path)
        except FileExistsError:
            pass
        except Exception as e:
            logging.info(f"Something went wrong while trying to create {Fore.CYAN}{folder_path}{Fore.RESET}: {e}")
        
    def _try_create_model_snapshot(self, type, chronicle_path):
        now = datetime.datetime.now()
        
        logging.info("Saving results...")
        
        if type == "naive":
            try:
                shutil.copyfile(f"{chronicle_path}\\weights\\best.pt", f"{chronicle_path}\\weights\\pre_{now.year}_{now.month}_{now.day}_{now.hour}{now.minute}{now.second}.pt")

                logging.info(f"Model saved at: {Fore.CYAN}{chronicle_path}\\weights\\pre_{now.year}_{now.month}_{now.day}_{now.hour}{now.minute}{now.second}.pt{Fore.RESET}")
            except Exception as e:
                logging.info(f"Couldn't create model snapshot, something went wrong: {e}")
        elif type == "wise":
            try:
                shutil.copyfile(f"{chronicle_path}\\weights\\best.pt", f"{chronicle_path}\\weights\\{now.year}_{now.month}_{now.day}_{now.hour}{now.minute}{now.second}.pt")

                logging.info(f"Model saved at: {Fore.CYAN}{chronicle_path}\\weights\\{now.year}_{now.month}_{now.day}_{now.hour}{now.minute}{now.second}.pt{Fore.RESET}")
            except Exception as e:
                logging.info(f"Couldn't create model snapshot, something went wrong: {e}")
        
    def _create_chronicle(self):
        raise NotImplementedError
        
    def train(self, epochs: int, *kwargs):
        raise NotImplementedError




import base64
import json
import logging
import os
import random
import re
import shutil

import albumentations as A

from colorama import Fore
import cv2
import yaml
from .dataset import BaseDatasetCommands
from .constants import README_CUSTOM_DATASETS


class DatasetCommandsDetection(BaseDatasetCommands):
    def __init__(self, dataset_name):
        self.type = "detection"
        self.dataset_name = dataset_name
        self.dataset_path = f"{os.path.abspath('.')}\\datasets\\{dataset_name}"
        
    def _create_dataset_metadata_files(self, **kwargs):
        self._try_create_folder(self.dataset_path)
        
        dataset_info = {"type": "obd", "version": "0.1.0", "snapshots": {}}
        
        with open(f"{self.dataset_path}\\dataset.json", "w", encoding="utf8") as wf:
            wf.write(json.dumps(dataset_info))

        with open(f"{self.dataset_path}\\snapshot.json", "w", encoding="utf8") as wf:
            wf.write(json.dumps({}))
            
        with open(
            f"{self.dataset_path}\\CUSTOM_DATASETS.md", "w", encoding="utf8"
        ) as wf:
            wf.write(README_CUSTOM_DATASETS)
            
        with open(f"{self.dataset_path}\\classes.txt", "w", encoding="utf8") as wf:
            wf.write("0: object")
            
        logging.info("Creating staging folders...")
        
        self._try_create_folder(self.dataset_path + "\\staging")
        self._try_create_folder(self.dataset_path + "\\staging\\images")
        self._try_create_folder(self.dataset_path + "\\staging\\labels")

        logging.info("Succesfully created staging folders.")

        self._try_create_folder(self.dataset_path)
        self._try_create_folder(self.dataset_path + "\\train")
        self._try_create_folder(self.dataset_path + "\\train\\images")
        self._try_create_folder(self.dataset_path + "\\train\\labels")
        self._try_create_folder(self.dataset_path + "\\val")
        self._try_create_folder(self.dataset_path + "\\val\\images")
        self._try_create_folder(self.dataset_path + "\\val\\labels")
        
        logging.info("Succesfully created dataset folders.")
        
        logging.info(
            f"Praise the gods! Your dataset folders are created, you can now insert your {Fore.CYAN}images{Fore.RESET} and {Fore.CYAN}labels{Fore.RESET} on {Fore.BLUE}YOLO{Fore.RESET} format at {Fore.CYAN}{self.dataset_path}\\staging{Fore.RESET} and then run {Fore.CYAN}odin dataset stage {self.dataset_name} --train=70 --val=30{Fore.RESET} (tip: you can change the {Fore.CYAN}--train{Fore.RESET} and {Fore.CYAN}--val{Fore.RESET} values to increase or decrease the split of the dataset)."
        )
        
    def _get_label_name_from_image_name(self, image_name="", **kwargs):
        pattern = r"\b(?:jpg|jpeg|png|webp|gif|bmp|tiff|svg)\b"
        return re.sub(pattern, "ext", image_name, flags=re.IGNORECASE)
        
    def _get_label_binary(self, label_path="", **kwargs):
        with open(label_path, "rb") as f:
            return f.read()
        
    def _add_artifact_to_version_snapshot(
        self, snapshot={}, dataset_split="", label_name="", label_binary="", image_name="", image_binary="", **kwargs
    ):
        try:
            snapshot[dataset_split]['images'].append(
                {
                    "filename": image_name,
                    "binary": base64.b64encode(image_binary).decode("utf8"),
                }
            )
        except:
            snapshot[dataset_split]['images'] = [
                {
                    "filename": image_name,
                    "binary": base64.b64encode(image_binary).decode("utf8"),
                }
            ]
            
        try:
            snapshot[dataset_split]['labels'].append(
                {
                    "filename": label_name,
                    "binary": base64.b64encode(label_binary).decode("utf8"),
                }
            )
        except:
            snapshot[dataset_split]['labels'] = [
                {
                    "filename": label_name,
                    "binary": base64.b64encode(label_binary).decode("utf8"),
                }
            ]
            
    def _execute_data_publishment(self, snapshot={}, split="", split_max_file=0, **kwargs):
        images = []
        labels = []

        for _, _, files in os.walk(f"{self.dataset_path}\\staging\\images"):
            images = files
            break

        for _, _, files in os.walk(f"{self.dataset_path}\\staging\\labels"):
            labels = files
            break
        
        for i in range(0, split_max_file + 1):
            try:
                image_stage_path = (
                    f"{self.dataset_path}\\staging\\images\\{images[i]}"
                )
                image_publish_path = (
                    f"{self.dataset_path}\\{split}\\images\\{images[i]}"
                )

                label_stage_path = (
                    f"{self.dataset_path}\\staging\\labels\\{labels[i]}"
                )
                label_publish_path = (
                    f"{self.dataset_path}\\{split}\\labels\\{labels[i]}"
                )

                shutil.move(
                    image_stage_path,
                    image_publish_path,
                )

                shutil.move(
                    label_stage_path,
                    label_publish_path,
                )
                
                image_binary = open(image_publish_path, "rb").read()
                label_binary = open(label_publish_path, "rb").read()
                
                self._add_artifact_to_version_snapshot(snapshot, split, labels[i], label_binary, images[i], image_binary)
            except IndexError:
                pass
            
    def _publish_data(self, update_type="", train=0, val=0, **kwargs):
        if (
            sum(
                len(files)
                for _, _, files in os.walk(f"{self.dataset_path}\\staging\\images")
            )
            == 0
        ):
            logging.info(
                f"The {Fore.CYAN}Staging{Fore.RESET} dataset is empty, so nothing will be published or updated."
            )
            return
        
        base_version = json.loads(
            open(f"{self.dataset_path}\\dataset.json", "r", encoding="utf8").read()
        )["version"]
        temp_version = self._upgrade_version(base_version, update_type)
        
        snapshot = {
            "staging": {},
            "train": {},
            "val": {},
        }
            
        logging.info("Publishing with the following splits:")
        logging.info(f"{Fore.CYAN}train{Fore.RESET}: {train}%")
        logging.info(f"{Fore.CYAN}val{Fore.RESET}: {val}%")
        
        count_train = int(
            (train / 100)
            * sum(
                len(files)
                for _, _, files in os.walk(f"{self.dataset_path}\\staging\\images")
            )
        )

        count_val = int(
            (val / 100)
            * sum(
                len(files)
                for _, _, files in os.walk(f"{self.dataset_path}\\staging\\images")
            )
        )
        
        self._execute_data_publishment(snapshot, "train", count_train)
        logging.info(
            f"Succesfully published {Fore.GREEN}train{Fore.RESET} data."
        )
        self._execute_data_publishment(snapshot, "val", count_val)
        logging.info(
            f"Succesfully published {Fore.GREEN}val{Fore.RESET} data."
        )
        
        dataset_info = json.loads(
            open(f"{self.dataset_path}\\dataset.json", "r", encoding="utf8").read()
        )
        snapshot_info = json.loads(
            open(f"{self.dataset_path}\\snapshot.json", "r", encoding="utf8").read()
        )

        snapshot_info[temp_version] = snapshot
        dataset_info["version"] = temp_version

        with open(f"{self.dataset_path}\\dataset.json", "w", encoding="utf8") as wf:
            wf.write(json.dumps(dataset_info, indent=4))
        logging.info(
            f"Succesfully updated dataset version to {Fore.CYAN}v{temp_version}{Fore.RESET}"
        )

        with open(f"{self.dataset_path}\\snapshot.json", "w", encoding="utf8") as wf:
            wf.write(json.dumps(snapshot_info, indent=4))
        logging.info(
            f"Succesfully registered snapshots for dataset {Fore.CYAN}v{temp_version}{Fore.RESET}"
        )
        
    def _status_sum_staging(self, **kwargs):
        return sum(
            len(files)
            for _, _, files in os.walk(
                f"{self.dataset_path}\\staging\\images"
            )
        )

    def _status_sum_train(self, **kwargs):
        return sum(
            len(files)
            for _, _, files in os.walk(
                f"{self.dataset_path}\\train\\images"
            )
        )
    
    def _status_sum_val(self, **kwargs):
        return sum(
            len(files)
            for _, _, files in os.walk(
                f"{self.dataset_path}\\val\\images"
            )
        )
        
    def _status_show_additional_info(self, **kwargs):
        train = self._status_sum_train()
        
        if train == 0:
            return

        with open(f"{self.dataset_path}\\data.yaml", "r", encoding="utf8") as f:
            dataset_yaml = yaml.safe_load(f)

        train_labels = []
        val_labels = []
        for _, _, files in os.walk(f"{self.dataset_path}\\train\\labels"):
            train_labels = files
        for _, _, files in os.walk(f"{self.dataset_path}\\val\\labels"):
            val_labels = files

        train_label_count = {}
        val_label_count = {}

        for label_file in train_labels:
            with open(
                f"{self.dataset_path}\\train\\labels\\{label_file}", "r", encoding="utf8"
            ) as f:
                data = f.read().split("\n")

                for line in data:
                    class_id = line.split(" ")[0]
                    try:
                        train_label_count[class_id] += 1
                    except:
                        train_label_count[class_id] = 1

        for label_file in val_labels:
            with open(
                f"{self.dataset_path}\\val\\labels\\{label_file}", "r", encoding="utf8"
            ) as f:
                data = f.read().split("\n")

                for line in data:
                    class_id = line.split(" ")[0]
                    try:
                        val_label_count[class_id] += 1
                    except:
                        val_label_count[class_id] = 1

        logging.info(f"Class count on {Fore.CYAN}train{Fore.RESET}:")
        for class_id in train_label_count:
            try:
                logging.info(
                    f"{dataset_yaml['names'][int(class_id)]}: {train_label_count[class_id]}"
                )
            except:
                pass

        logging.info(f"Class count on {Fore.CYAN}val{Fore.RESET}:")
        for class_id in val_label_count:
            try:
                logging.info(
                    f"{dataset_yaml['names'][int(class_id)]}: {val_label_count[class_id]}"
                )
            except:
                pass
            
    def _augmentate_data(self, augmentation_amount=0, **kwargs):
        try:
            try:
                with open(
                    f"{self.dataset_path}\\data.yaml", "r", encoding="utf8"
                ) as data:
                    data_loaded = yaml.safe_load(data)
            except:
                logging.info(
                    f"The dataset's {Fore.CYAN}data.yaml{Fore.RESET} wasn't found. Create one by running the command {Fore.CYAN}odin dataset yaml {self.dataset_name}{Fore.RESET}"
                )
                return

            images = []
            for _, _, files in os.walk(f"{self.dataset_path}\\train\\images"):
                images = files
                break

            logging.info(
                f"Augmentating {Fore.CYAN}{len(images)}{Fore.RESET} images to a total of {Fore.CYAN}{len(images)+(len(images)*augmentation_amount)}{Fore.RESET} images..."
            )
            for image_file in images:
                image = cv2.imread(f"{self.dataset_path}\\train\\images\\{image_file}")

                image_id = (
                    image_file.split(".png")[0].split(".jpg")[0].split(".jpeg")[0]
                )

                image_height, image_width, image_channels = image.shape
                image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

                bboxes = list(
                    map(
                        lambda y: (
                            [
                                int((float(y[1]) - float(y[3]) / 2) * image_width),
                                int((float(y[2]) - float(y[4]) / 2) * image_height),
                                float(y[3]) * image_width,
                                float(y[4]) * image_height,
                            ]
                            if len(y) > 1
                            else None
                        ),
                        list(
                            map(
                                lambda x: x.split(" "),
                                open(
                                    f"{self.dataset_path}\\train\\labels\\{image_id}.txt"
                                )
                                .read()
                                .split("\n"),
                            )
                        ),
                    )
                )
                try:
                    bboxes.remove(None)
                except:
                    pass
                classes = list(
                    map(
                        lambda y: y[0] if len(y) > 1 else None,
                        list(
                            map(
                                lambda x: x.split(" "),
                                open(
                                    f"{self.dataset_path}\\train\\labels\\{image_id}.txt"
                                )
                                .read()
                                .split("\n"),
                            )
                        ),
                    )
                )
                try:
                    classes.remove(None)
                except:
                    pass
                classes = list(map(lambda z: int(z), classes))  # type: ignore
                classes_to_name = data_loaded["names"]

                transform = A.Compose(
                    [
                        A.HorizontalFlip(p=0.5),
                        # A.ShiftScaleRotate(p=0.5),
                        A.RandomCropFromBorders(p=0.5),
                        A.RandomBrightnessContrast(p=0.5),
                        A.RGBShift(
                            r_shift_limit=(-50, 50),
                            g_shift_limit=(-50, 50),
                            b_shift_limit=(-50, 50),
                            p=0.5,
                        ),
                        A.CLAHE(p=0.5),
                        A.ISONoise(intensity=(0.3, 0.5), p=0.5),
                        A.RandomGravel(p=0.5),
                        A.HueSaturationValue(p=0.5),
                    ],
                    bbox_params=A.BboxParams(
                        format="coco", label_fields=["classes"]
                    ),
                )

                random.seed(7)
                data_to_save = []

                for i in range(0, augmentation_amount):
                    data_to_save.append(
                        transform(image=image, bboxes=bboxes, classes=classes)
                    )

                def get_yolo_bboxes(bboxes):
                    final_bboxes = []
                    for bbox in bboxes:
                        x_min, y_min, w, h = bbox

                        x_center = (x_min + w / 2) / image_width
                        y_center = (y_min + h / 2) / image_height
                        width = w / image_width
                        height = h / image_height

                        final_bboxes.append([x_center, y_center, width, height])
                    return final_bboxes

                def get_yolo_label(classes, bboxes):
                    lines = []
                    for i in range(0, len(classes)):
                        bbox = list(map(lambda x: str(x), bboxes[i]))
                        bboxes_lines = " ".join(bbox)
                        lines.append(f"{classes[i]} {bboxes_lines}")
                    return "\n".join(lines)

                annotation_id = 0
                for data in data_to_save:
                    image = cv2.cvtColor(data["image"], cv2.COLOR_BGR2RGB)
                    cv2.imwrite(
                        f"{self.dataset_path}\\train\\images\\{image_id}-{annotation_id}.png",
                        image,
                    )

                    bboxes = get_yolo_bboxes(data["bboxes"])
                    classes = list(map(lambda x: int(x), data["classes"]))
                    yolo_label = get_yolo_label(classes, bboxes)

                    with open(
                        f"{self.dataset_path}\\train\\labels\\{image_id}-{annotation_id}.txt",
                        "w",
                    ) as wf:
                        wf.write(yolo_label)

                    annotation_id += 1

            logging.info("Succesfully augmented all images.")
        except Exception as e:
            print(e)
            pass

    def _rollback_dataset(self, snapshot_info={}, rollback_version="", dataset_folder="", split_folder="", **kwargs):
        shutil.rmtree(f"{self.dataset_path}\\{dataset_folder}\\{split_folder}\\images")
        os.makedirs(f"{self.dataset_path}\\{dataset_folder}\\{split_folder}\\images")
        shutil.rmtree(f"{self.dataset_path}\\{dataset_folder}\\{split_folder}\\labels")
        os.makedirs(f"{self.dataset_path}\\{dataset_folder}\\{split_folder}\\labels")    
        
        for file_data in snapshot_info[rollback_version][dataset_folder][split_folder]:
            file_bin = file_data["binary"]
            file_name = file_data["filename"]

            encoded_content = file_bin.encode()
            content = base64.b64decode(encoded_content)

            with open(
                f"{self.dataset_path}\\{dataset_folder}\\{split_folder}\\{file_name}",
                "wb",
            ) as wf:
                wf.write(content)
        
        logging.info(
            f"Succesfully executed rollback at {Fore.CYAN}{dataset_folder}{Fore.RESET}"
        )

    def yaml(self, **kwargs):
        dataset_path = f"{os.path.abspath('.')}\\datasets\\{self.dataset_name}"

        if not os.path.exists(dataset_path):
            logging.info(
                f"The dataset mentioned doesn't exist, create it by using the command {Fore.CYAN}odin dataset create {self.dataset_name}{Fore.RESET}."
            )
        else:
            try:
                with open(f"{dataset_path}\\classes.txt", "r", encoding="utf8") as f:
                    data = list(
                        map(
                            lambda x: x.split(":")[-1].replace(" ", ""),
                            f.read().split("\n"),
                        )
                    )

                dataset_yaml = {
                    "path": dataset_path,
                    "train": "train\\images",
                    "val": "val\\images",
                    "names": {},
                }

                for class_id in range(0, len(data)):
                    dataset_yaml["names"][class_id] = data[class_id]

                with open(f"{dataset_path}\\data.yaml", "w", encoding="utf8") as wf:
                    wf.write(yaml.dump(dataset_yaml))

                logging.info(
                    f"Succesfully generated {Fore.CYAN}data.yaml{Fore.RESET} for {Fore.CYAN}{self.dataset_name}{Fore.RESET}"
                )
            except:
                logging.info(
                    f"Your {Fore.CYAN}classes.txt{Fore.RESET} is either empty or non-existant. If you don't have a {Fore.CYAN}classes.txt{Fore.RESET} in your dataset, please provide one so {Fore.CYAN}Odin{Fore.RESET} can generate the YAML file. Read the documentation to know more about how it should be at {Fore.CYAN}datasets/{self.dataset_name}/CUSTOM_DATASETS.md{Fore.RESET}."
                )
            


import base64
import json
import logging
import os
import random
import shutil

import click
import albumentations as A
from colorama import Fore
import cv2

from .constants import README_CUSTOM_DATASETS
from .dataset import BaseDatasetCommands


class DatasetCommandsClassification(BaseDatasetCommands):
    def __init__(self, dataset_name):
        self.type = "classification"
        self.dataset_name = dataset_name
        self.dataset_path = f"{os.path.abspath('.')}\\datasets\\{dataset_name}"

    def _create_dataset_metadata_files(self, **kwargs):
        os.makedirs(self.dataset_path)

        dataset_info = {
            "type": "classification",
            "version": "0.1.0",
            "snapshots": {},
        }

        with open(f"{self.dataset_path}\\dataset.json", "w", encoding="utf8") as wf:
            wf.write(json.dumps(dataset_info))

        with open(f"{self.dataset_path}\\snapshot.json", "w", encoding="utf8") as wf:
            wf.write(json.dumps({}))

        with open(
            f"{self.dataset_path}\\CUSTOM_DATASETS.md", "w", encoding="utf8"
        ) as wf:
            wf.write(README_CUSTOM_DATASETS)

        os.makedirs(self.dataset_path + "\\train")
        os.makedirs(self.dataset_path + "\\val")

        logging.info("Succesfully created final folders.")

        # Creating dataset staging folder

        logging.info("Creating staging folders...")

        os.makedirs(self.dataset_path + "\\staging")
        os.makedirs(self.dataset_path + "\\staging\\class_1")
        os.makedirs(self.dataset_path + "\\staging\\class_2")

        logging.info("Succesfully created staging folders.")

        # Praise the gods! Your dataset folders are created, you can now insert your images and labels on YOLO format at 'self.dataset_path\\staging' and then run 'odin dataset stage {dataset_name} --train=70 --val=30'
        logging.info(
            f"Praise the gods! Your dataset folders are created, you can now insert your {Fore.CYAN}images{Fore.RESET} at {Fore.CYAN}{self.dataset_path}\\staging{Fore.RESET} and then run {Fore.CYAN}odin dataset stage {self.dataset_name} --train=70 --val=30{Fore.RESET} (tip: you can change the {Fore.CYAN}--train{Fore.RESET} and {Fore.CYAN}--val{Fore.RESET} values to increase or decrease the split of the dataset)."
        )
        logging.info(
            f"{Fore.GREEN}NOTE{Fore.RESET}: {Fore.CYAN}Odin{Fore.RESET} created two temporary classes as placeholders to your classification project called {Fore.CYAN}class_1{Fore.RESET} and {Fore.CYAN}class_2{Fore.RESET}. You can delete them and include your own classes. If you don't know how to create a classification dataset follow the Ultralytics guide available at: https://docs.ultralytics.com/datasets/classify/#dataset-structure-for-yolo-classification-tasks"
        )

    def _add_image_to_version_snapshot(
        self, snapshot={}, dataset_split="", dataset_class="", image_name="", image_binary="", **kwargs
    ):
        try:
            snapshot[dataset_split][dataset_class].append(
                {
                    "filename": image_name,
                    "binary": base64.b64encode(image_binary).decode("utf8"),
                }
            )
        except:
            snapshot[dataset_split][dataset_class] = [
                {
                    "filename": image_name,
                    "binary": base64.b64encode(image_binary).decode("utf8"),
                }
            ]

    def _execute_data_publishment(self, snapshot={}, split="", dataset_class="", split_max_file=0, **kwargs):
        images = []

        for _, _, files in os.walk(f"{self.dataset_path}\\staging\\{dataset_class}"):
            images = files
            break

        for i in range(0, split_max_file + 1):
            try:
                image_stage_path = (
                    f"{self.dataset_path}\\staging\\{dataset_class}\\{images[i]}"
                )
                
                self._try_create_folder(f"{self.dataset_path}\\{split}\\{dataset_class}")
                
                image_publish_path = (
                    f"{self.dataset_path}\\{split}\\{dataset_class}\\{images[i]}"
                )

                shutil.move(
                    image_stage_path,
                    image_publish_path,
                )

                image_binary = open(image_publish_path, "rb").read()

                self._add_image_to_version_snapshot(
                    snapshot, split, dataset_class, images[i], image_binary
                )
            except IndexError:
                pass

    def _publish_data(self, update_type="", train=0, val=0, *args):
        classes = []

        base_version = json.loads(
            open(f"{self.dataset_path}\\dataset.json", "r", encoding="utf8").read()
        )["version"]
        temp_version = self._upgrade_version(base_version, update_type)

        # SNAPSHOT UPDATE
        snapshot = {
            "staging": {},
            "train": {},
            "val": {},
        }

        for x in os.walk(f"{self.dataset_path}\\staging"):
            if len(x[1]) > 0:
                classes = x[1]

        for dataset_class in classes:
            if (
                sum(
                    len(files)
                    for _, _, files in os.walk(
                        f"{self.dataset_path}\\staging\\{dataset_class}"
                    )
                )
                == 0
            ):
                logging.info(
                    f"The {Fore.CYAN}Staging{Fore.RESET} dataset is empty, so nothing will be published or updated."
                )
                return

            logging.info(f"Publishing {Fore.CYAN}{dataset_class}{Fore.RESET}")

            if (
                sum(
                    len(files)
                    for _, _, files in os.walk(
                        f"{self.dataset_path}\\staging\\{dataset_class}"
                    )
                )
                == 0
            ):
                logging.info(
                    f"The {Fore.CYAN}Staging{Fore.RESET} dataset for class {Fore.CYAN}{dataset_class}{Fore.RESET} is empty, so nothing will be published or updated for this class. Skipping this one."
                )
                return

            logging.info("Publishing with the following splits:")
            logging.info(f"{Fore.CYAN}train{Fore.RESET}: {train}%")
            logging.info(f"{Fore.CYAN}val{Fore.RESET}: {val}%")

            count_train = int(
                (train / 100)
                * sum(
                    len(files)
                    for _, _, files in os.walk(
                        f"{self.dataset_path}\\staging\\{dataset_class}"
                    )
                )
            )

            count_val = int(
                (val / 100)
                * sum(
                    len(files)
                    for _, _, files in os.walk(
                        f"{self.dataset_path}\\staging\\{dataset_class}"
                    )
                )
            )

            self._execute_data_publishment(
                snapshot, "train", dataset_class, count_train
            )
            logging.info(
                f"Succesfully published {Fore.GREEN}train{Fore.RESET} data for class {Fore.CYAN}{dataset_class}{Fore.RESET}"
            )
            self._execute_data_publishment(snapshot, "val", dataset_class, count_val)
            logging.info(
                f"Succesfully published {Fore.GREEN}val{Fore.RESET} data for class {Fore.CYAN}{dataset_class}{Fore.RESET}"
            )

        dataset_info = json.loads(
            open(f"{self.dataset_path}\\dataset.json", "r", encoding="utf8").read()
        )
        snapshot_info = json.loads(
            open(f"{self.dataset_path}\\snapshot.json", "r", encoding="utf8").read()
        )

        snapshot_info[temp_version] = snapshot
        dataset_info["version"] = temp_version

        with open(f"{self.dataset_path}\\dataset.json", "w", encoding="utf8") as wf:
            wf.write(json.dumps(dataset_info, indent=4))
        logging.info(
            f"Succesfully updated dataset version to {Fore.CYAN}v{temp_version}{Fore.RESET}"
        )

        with open(f"{self.dataset_path}\\snapshot.json", "w", encoding="utf8") as wf:
            wf.write(json.dumps(snapshot_info, indent=4))
        logging.info(
            f"Succesfully registered snapshots for dataset {Fore.CYAN}v{temp_version}{Fore.RESET}"
        )

    def _status_sum_staging(self, **kwargs):
        classes = []
        for x in os.walk(f"{self.dataset_path}\\staging"):
            if len(x[1]) > 0:
                classes = x[1]

        final_sum = 0
        for dataset_class in classes:
            final_sum += sum(
                len(files)
                for _, _, files in os.walk(
                    f"{self.dataset_path}\\staging\\{dataset_class}"
                )
            )

        return final_sum

    def _status_sum_train(self, **kwargs):
        classes = []
        for x in os.walk(f"{self.dataset_path}\\staging"):
            if len(x[1]) > 0:
                classes = x[1]

        final_sum = 0
        for dataset_class in classes:
            final_sum += sum(
                len(files)
                for _, _, files in os.walk(
                    f"{self.dataset_path}\\train\\{dataset_class}"
                )
            )

        return final_sum

    def _augmentate_data(self, augmentation_amount=0, **kwargs):
        try:
            classes = []
            for x in os.walk(f"{self.dataset_path}\\staging"):
                if len(x[1]) > 0:
                    classes = x[1]

            for dataset_class in classes:
                images = []
                for _, _, files in os.walk(
                    f"{self.dataset_path}\\train\\{dataset_class}"
                ):
                    images = files
                    break

                if len(images) == 0:
                    logging.info(
                        f"There are no images for {Fore.CYAN}{dataset_class}{Fore.RESET}, skipping the augmentation process for this class."
                    )
                    return

                logging.info(
                    f"Augmentating {Fore.CYAN}{dataset_class}{Fore.RESET} {Fore.CYAN}{len(images)}{Fore.RESET} images to a total of {Fore.CYAN}{len(images)+(len(images)*augmentation_amount)}{Fore.RESET} images..."
                )
                for image_file in images:
                    image = cv2.imread(
                        f"{self.dataset_path}\\train\\{dataset_class}\\{image_file}"
                    )

                    image_id = (
                        image_file.split(".png")[0].split(".jpg")[0].split(".jpeg")[0]
                    )

                    image_height, image_width, image_channels = image.shape
                    image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

                    transform = A.Compose(
                        [
                            A.HorizontalFlip(p=0.5),
                            # A.ShiftScaleRotate(p=0.5),
                            A.RandomCropFromBorders(p=0.5),
                            A.RandomBrightnessContrast(p=0.5),
                            A.RGBShift(
                                r_shift_limit=(-50, 50),
                                g_shift_limit=(-50, 50),
                                b_shift_limit=(-50, 50),
                                p=0.5,
                            ),
                            A.CLAHE(p=0.5),
                            A.ISONoise(intensity=(0.3, 0.5), p=0.5),
                            A.RandomGravel(p=0.5),
                            A.HueSaturationValue(p=0.5),
                        ]
                    )

                    random.seed(7)
                    data_to_save = []

                    for i in range(0, augmentation_amount):
                        data_to_save.append(transform(image=image))

                    annotation_id = 0
                    for data in data_to_save:
                        image = cv2.cvtColor(data["image"], cv2.COLOR_BGR2RGB)
                        cv2.imwrite(
                            f"{self.dataset_path}\\train\\{dataset_class}\\{image_id}-{annotation_id}.png",
                            image,
                        )

                        annotation_id += 1

            logging.info("Succesfully augmented all images.")
        except:
            pass

    def _rollback_dataset(
        self, snapshot_info={}, rollback_version="", dataset_folder="", class_folder="", **kwargs
    ):
        shutil.rmtree(f"{self.dataset_path}\\{dataset_folder}\\{class_folder}")
        os.makedirs(f"{self.dataset_path}\\{dataset_folder}\\{class_folder}")

        for file_data in snapshot_info[rollback_version][dataset_folder][class_folder]:
            img_bin = file_data["binary"]
            img_name = file_data["filename"]

            encoded_content = img_bin.encode()
            content = base64.b64decode(encoded_content)

            with open(
                f"{self.dataset_path}\\{dataset_folder}\\{class_folder}\\{img_name}",
                "wb",
            ) as wf:
                wf.write(content)

        logging.info(
            f"Succesfully executed rollback at {Fore.CYAN}{dataset_folder}{Fore.RESET}"
        )

    def _status_show_additional_info(self, **kwargs):
        classes = []
        for x in os.walk(f"{self.dataset_path}\\staging"):
            if len(x[1]) > 0:
                classes = x[1]

        staging = self._status_sum_staging()
        train = self._status_sum_train()

        train_images_count = 0
        val_images_count = 0
        # Image Count
        for dataset_class in classes:
            train_images_count += sum(
                len(files)
                for _, _, files in os.walk(
                    f"{self.dataset_path}\\train\\{dataset_class}"
                )
            )
            val_images_count += sum(
                len(files)
                for _, _, files in os.walk(f"{self.dataset_path}\\val\\{dataset_class}")
            )

        logging.info(f"Images on {Fore.CYAN}train{Fore.RESET}: {train_images_count}")
        logging.info(f"Images on {Fore.CYAN}val{Fore.RESET}: {val_images_count}")

        # Class count
        if train == 0:
            return

        train_count_info = {}
        val_count_info = {}

        for dataset_class in classes:
            train_count_info[dataset_class] = 0
            val_count_info[dataset_class] = 0

        for dataset_class in classes:
            train_count_info[dataset_class] += sum(
                len(files)
                for _, _, files in os.walk(
                    f"{self.dataset_path}\\train\\{dataset_class}"
                )
            )
            val_count_info[dataset_class] += sum(
                len(files)
                for _, _, files in os.walk(f"{self.dataset_path}\\val\\{dataset_class}")
            )

        logging.info("")
        logging.info(f"Class count on {Fore.CYAN}train{Fore.RESET}:")
        logging.info("")
        for class_id in train_count_info:
            try:
                if train_count_info[class_id] == 0:
                    logging.info(
                        f"{Fore.CYAN}{class_id}{Fore.RESET}: {Fore.LIGHTRED_EX}{train_count_info[class_id]}{Fore.RESET}"
                    )
                else:
                    logging.info(
                        f"{Fore.CYAN}{class_id}{Fore.RESET}: {train_count_info[class_id]}"
                    )
            except:
                pass

        logging.info("")
        logging.info(f"Class count on {Fore.CYAN}val{Fore.RESET}:")
        logging.info("")
        for class_id in val_count_info:
            try:
                if val_count_info[class_id] == 0:
                    logging.info(
                        f"{Fore.CYAN}{class_id}{Fore.RESET}: {Fore.LIGHTRED_EX}{val_count_info[class_id]}{Fore.RESET}"
                    )
                else:
                    logging.info(
                        f"{Fore.CYAN}{class_id}{Fore.RESET}: {val_count_info[class_id]}"
                    )
            except:
                pass

    def yaml(self, **kwargs):
        logging.info(
            f"Your dataset type is defined as {Fore.CYAN}Classification{Fore.RESET}, this type of project doesn't require a {Fore.CYAN}data.yaml{Fore.RESET}."
        )


import os
import yaml


def get_project_info():
    path = os.path.abspath(".") + "\\project.yaml"

    with open(path, "r") as f:
        data = yaml.safe_load(f)

    return data


def get_project_pretty_type(type):
    types = {"detection": "Object Detection", "classification": "Image Classification"}
    return types[type]




import cv2


class YOLOResult:
    def __init__(self, image_with_bboxes, inference_speed):
        self.image_with_bboxes = image_with_bboxes
        self.inference_speed = inference_speed
        
    def show(self):
        cv2.imshow("Inference Result", self.image_with_bboxes)
        cv2.waitKey(0)
        cv2.destroyAllWindows()



import datetime
import json
import logging
import os
import shutil

import click
from colorama import Fore
from ultralytics import YOLO
from .chronicle_utils import get_chronicle_name
from .training import BaseTrainingCommands


class ClassificationTrainingCommands(BaseTrainingCommands):
    def train(self, epochs: int=0, device: str="", base_model: str="", dataset_name: str="", chronicle_name: str="", subset: int=0, **kwargs):
        if not base_model:
            base_model = "yolo11n-cls.pt"
        
        if subset < 100:
            model_type = "naive"
        else:
            model_type = "wise"
            
        if not dataset_name:
            confirmed_name = False
            
            while not confirmed_name:
                dataset_name = click.prompt(
                    f"What is the dataset's name?"
                )
                
                if click.confirm(f'{Fore.CYAN}{dataset_name}{Fore.RESET}, is this name right?'):
                    confirmed_name = True
        
        if not chronicle_name:
            chronicle_name = click.prompt(f"What will be the name of the {Fore.CYAN}chronicle{Fore.RESET}? If left empty, will create at", default=get_chronicle_name())
        
        chronicles_path = f"{os.path.abspath('.')}\\chronicles" 
        chronicle_path = f"{os.path.abspath('.')}\\chronicles\\{chronicle_name}"
        
        dataset_path = f"{os.path.abspath('.')}\\datasets\\{dataset_name}"
        
        logging.info(f"Starting {Fore.CYAN}training{Fore.RESET}...")
        logging.info(f"Version defined as {Fore.CYAN}{model_type}{Fore.RESET}")
        
        
        self._try_create_folder(chronicle_path)
        self._try_create_folder(chronicle_path+"\\weights")
        
        chronicle_info = {
            "name": chronicle_name,
            "dataset": dataset_name
        }
        
        with open(chronicle_path+"\\chronicle.json", "w", encoding="utf8") as wf:
            wf.write(json.dumps(chronicle_info, indent=4, ensure_ascii=False))
        
        yolo_command = [
            "yolo",
            "classify",
            "train",
            f"data={dataset_path}",
            f"epochs={epochs}",
            f"batch={'-1' if device == 'gpu' else '2'}",
            f"model={base_model}",
            "amp=false",
            "patience=10",
            "save_period=5",
            f"device={'0' if device == 'gpu' else 'cpu'}",
            f"project={chronicles_path}",
            f"name={chronicle_name}",
            "exist_ok=true",
            "plots=true",
        ]
        
        if model_type == "naive":
            yolo_command.append(f"fraction={subset/100}")
        
        os.system(
            " ".join(yolo_command)
        )
        
        self._try_create_model_snapshot(model_type, chronicle_path)
        
        logging.info(f"Trained to chronicle {Fore.CYAN}{chronicle_name}{Fore.RESET}")
        logging.info(f"You can test this version by using the command {Fore.CYAN}odin test --chronicle {chronicle_name}{Fore.RESET}")
        
    def test(self, chronicle_name: str="", **kwargs):
        if not chronicle_name:
            confirmed_name = False
            
            while not confirmed_name:
                chronicle_name = click.prompt(
                    f"What is the chronicle's name?"
                )
                
                if click.confirm(f'{Fore.CYAN}{chronicle_name}{Fore.RESET}, is this name right?'):
                    confirmed_name = True
        
        chronicle_data = self._get_chronicle_data(f"{os.path.abspath('.')}\\chronicles\\{chronicle_name}")
        dataset_name = chronicle_data["dataset"]
        
        classes = []
        for _, class_folders, _ in os.walk(f"{os.path.abspath('.')}\\datasets\\{dataset_name}\\val"):
            classes = class_folders
            break
        
        chronicle_models_path = f"{os.path.abspath('.')}\\chronicles\\{chronicle_name}\\weights"
        eligible_models = []
        
        for _, _, files in os.walk(chronicle_models_path):
            for file in files:
                if self._check_if_eligible_model(file):
                    eligible_models.append(file)
            
        eligible_models.sort(reverse=True)
            
        model_selection = {}
        model_id = 0
        model_id_options = []
        for model in eligible_models:
            model_selection[str(model_id)] = model
            model_id_options.append(str(model_id))
            model_id += 1
        
        print("")
        for model_selection_id in model_selection:
            print(f"{Fore.CYAN}{model_selection_id}{Fore.RESET} - {model_selection[model_selection_id]}")
        print("")
        
        model_selected = model_selection[click.prompt("Select a model to test", default="0", show_choices=True, type=click.Choice(model_id_options))]
        print("")
        
        model: YOLO = YOLO(f"{os.path.abspath('.')}\\chronicles\\{chronicle_name}\\weights\\{model_selected}", verbose=False)
                    
        logging.info(f"Successfully loaded model {Fore.CYAN}{model_selected}{Fore.RESET}.")
        
        for dataset_class in classes:
            logging.info(f"Testing the following class: {Fore.CYAN}{dataset_class}{Fore.RESET}")
            
            val_images_path = f"{os.path.abspath('.')}\\datasets\\{dataset_name}\\val\\{dataset_class}"
            
            val_images = []
            
            for _, _, files in os.walk(val_images_path):
                for file in files:
                    image_path = f"{os.path.abspath('.')}\\datasets\\{dataset_name}\\val\\{dataset_class}\\{file}"
                    val_images.append(image_path)
            
            logging.info(f"Loading model {Fore.CYAN}{model_selected}{Fore.RESET}...")
                    
            results = model.predict(val_images, verbose=False)
            odin_results = self._gather_yolo_results(results)
                
            for result in odin_results:
                logging.info(f"Speed: {result.inference_speed}")
                result.show()
                
    def publish(self, project_name: str="", chronicle_name: str="", **kwargs):
        if not chronicle_name:
            confirmed_name = False
            
            while not confirmed_name:
                chronicle_name = click.prompt(
                    f"What is the chronicle's name?"
                )
                
                if click.confirm(f'{Fore.CYAN}{chronicle_name}{Fore.RESET}, is this name right?'):
                    confirmed_name = True
        
        chronicle_data = self._get_chronicle_data(f"{os.path.abspath('.')}\\chronicles\\{chronicle_name}")
        
        chronicle_models_path = f"{os.path.abspath('.')}\\chronicles\\{chronicle_name}\\weights"
        eligible_models = []
        
        for _, _, files in os.walk(chronicle_models_path):
            for file in files:
                if self._check_if_eligible_model(file):
                    eligible_models.append(file)
            
        eligible_models.sort(reverse=True)
            
        model_selection = {}
        model_id = 0
        model_id_options = []
        for model in eligible_models:
            model_selection[str(model_id)] = model
            model_id_options.append(str(model_id))
            model_id += 1
        
        print("")
        for model_selection_id in model_selection:
            print(f"{Fore.CYAN}{model_selection_id}{Fore.RESET} - {model_selection[model_selection_id]}")
        print("")
        
        model_selected = model_selection[click.prompt("Select a model to test", default="0", show_choices=True, type=click.Choice(model_id_options))]
        print("")
        
        logging.info(f"Publishing model {Fore.CYAN}{model_selected}{Fore.RESET}...")
        
        now = datetime.datetime.now()
        new_model_name = f"{project_name}_{now.year}_{now.month}_{now.day}_{chronicle_name}.pt"
        
        shutil.copy(f"{os.path.abspath('.')}\\chronicles\\{chronicle_name}\\weights\\{model_selected}", f"{os.path.abspath('.')}\\weights\\{new_model_name}")
        
        logging.info(f"Succesfully published model {Fore.CYAN}{model_selected}{Fore.RESET} as {Fore.CYAN}{new_model_name}{Fore.RESET}...")


import os


EXAMPLE_YAML = f"""
path: {os.path.abspath('.')}\\datasets\\dataset_1_odb
train: train\\images
val: val\\images

names:
    0: dog
    1: cat
"""

README_CHRONICLES = """
# Odin's Chronicles

Chronicles are essentially training sessions that contain runs. Each sessions uses a specific dataset, so a good practice is to use/create a new session for each new version of the project's dataset.

## Case Example

A simple example of a chronicle would be: 

```
{project_path}/chronicles/my-chronicle/run-id
# Which contains:
.../my-chronicle/run-id/weights/best.pt
.../my-chronicle/run-id/labels.jpg
.../my-chronicle/run-id/results.csv
.../my-chronicle/run-id/train_batch0.jpg
...
```
"""

README_WEIGHTS = """
# Odin's Weights

Weights are the final versions of a model training. This is nothing but a way to organize your project, in fact, the file stored here is the same file acquired at `chronicles/my-chronicle/run-id/weights`.

This folder is managed by Odin and should not be changed. The framework itself versions the weights.
"""

README_CUSTOM_DATASETS = """
# Custom Datasets on Odin

## Object Detection

All **Object Detection** datasets created by **Odin** have a **classes.txt** inside it, this file is extremely important if you don't know how to create a **data.yaml** (if you know, you can just create your own **data.yaml** without any problems, but it's advised to let **Odin** do it, so the path is not wrong).

### classes.txt

The file **classes.txt** should be in the following format:

```
0: dog
1: cat
2: person
3: chair
```

## Classification

"""
